# TPK4186 - Advanced Tools for Performance Engineering
# ProcessMonitoringGenerator.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required Modules
# 2. Monitored Process
# 3. Sample Generation

# 1. Required Modules
# -------------------

import sys
import random
import numpy

# 2. Monitored Process
# --------------------

class MonitoredProcess:
  def __init__(self):
    self.version = 1
    self.low1 = 0.0
    self.high1 = 1.0
    self.low2 = 0.0
    self.high2 = 1.0
    self.threshold = 1.0

  def KOutOfN(self, Ps, k):
    Rs = [0.0 for _ in range(0, k+1)]
    Rs[0] = 1
    for i in range(0, len(Ps)):
      Rs[k] +=  Ps[i]*Rs[k-1]
      for j in range(k-1, 0, -1):
        Rs[j] = Ps[i]*Rs[j-1] + (1-Ps[i])*Rs[j]
      Rs[0] = (1-Ps[i])*Rs[0]
    return Rs[k]

  def CalculateRewardOfConfiguration(self, Cs):
    A = self.KOutOfN(Cs[0:2], 2)
    B = self.KOutOfN(Cs[2:4], 2)
    C = self.KOutOfN(Cs[4:6], 2)
    E = self.KOutOfN([A, B, C], 2)
    D = self.KOutOfN(Cs[6:10], 3)
    R = self.KOutOfN([E, D], 1)
    return R

  def GenerateConfiguration(self):
    A1 = random.uniform(self.low1, self.high1)
    A2 = random.uniform(self.low1, self.high1)
    B1 = random.uniform(self.low1, self.high1)
    B2 = random.uniform(self.low1, self.high1)
    C1 = random.uniform(self.low1, self.high1)
    C2 = random.uniform(self.low1, self.high1)
    D1 = random.uniform(self.low2, self.high2)
    D2 = random.uniform(self.low2, self.high2)
    D3 = random.uniform(self.low2, self.high2)
    D4 = random.uniform(self.low2, self.high2)
    return [A1, A2, B1, B2, C1, C2, D1, D2, D3, D4]

  def GenerateSampleForRegression(self, numberOfElements):
    sample = numpy.zeros((numberOfElements, 10))
    rewards = numpy.zeros(numberOfElements)
    for index in range(0, numberOfElements):
      Cs = self.GenerateConfiguration()
      reward = self.CalculateRewardOfConfiguration(Cs)
      sample[index] = Cs
      rewards[index] = reward
    return (sample, rewards)

  def GenerateSampleForClassification(self, numberOfElements):
    sample = numpy.zeros((numberOfElements, 10))
    labels = numpy.zeros(numberOfElements)
    for index in range(0, numberOfElements):
      Cs = self.GenerateConfiguration()
      reward = self.CalculateRewardOfConfiguration(Cs)
      sample[index] = Cs
      if reward>self.threshold:
        labels[index] = 1
      else:
        labels[index] = 0
    return (sample, labels)

  def GenerateReducedSampleForClassification(self, numberOfElements):
    sample = numpy.zeros((numberOfElements, 7))
    labels = numpy.zeros(numberOfElements)
    for index in range(0, numberOfElements):
      Cs = self.GenerateConfiguration()
      reward = self.CalculateRewardOfConfiguration(Cs)
      Rs = []
      for i in range(0, 6):
        Rs.append(Cs[i])
      Rs.append(self.KOutOfN(Cs[6:10], 3))
      sample[index] = Rs
      if reward>self.threshold:
        labels[index] = 1
      else:
        labels[index] = 0
    return (sample, labels)

  def PrintSample(self, sample, rewards, fileName, separator):
    try:
      file = open(fileName, "w")
    except:
      sys.stderr.write("Unable to open file " + fileName + "\n")
      sys.stderr.flush()
      return
    for row in range(0, len(sample)):
      point = sample[row]
      for attribute in point:
        file.write("{0:f}\t".format(attribute))
      file.write("{0:f}\n".format(rewards[row]))
    file.flush()
    file.close()

  def PrintStatisticsForRegression(self, rewards):
    print("Reward {0:.4e}:\t{1:.4e}\t{2:.4e}".format(rewards.min(), rewards.max(), rewards.mean()))
    print(numpy.quantile(rewards, [0.50, 0.90, 0.95]))

  def PrintStatisticsForClassification(self, labels):
    counts = [0, 0]
    for i in range(0, len(labels)):
      counts[int(labels[i])] += 1
    for c in range(0, 2):
      print("Label {0:d}:\t{1:d}\t{2:.1f}".format(c, counts[c], 100*counts[c]/len(labels)))

# 3. Sample Generation
# --------------------

random.seed(12345)
monitoredProcess = MonitoredProcess()
monitoredProcess.version = 7
monitoredProcess.low1 = 0.1
monitoredProcess.high1 = 0.2
monitoredProcess.low2 = 0.05
monitoredProcess.high2 = 0.10
monitoredProcess.threshold = 0.00305
#monitoredProcess.threshold = 0.00395
#monitoredProcess.threshold = 0.00480

trainingSampleSize = 10000
(trainingSample, trainingLabels) = monitoredProcess.GenerateReducedSampleForClassification(trainingSampleSize)
monitoredProcess.PrintSample(trainingSample, trainingLabels, "trainingSet{0:d}.csv".format(monitoredProcess.version), "\t")
monitoredProcess.PrintStatisticsForClassification(trainingLabels)

testSampleSize = 10000
(testSample, testLabels) = monitoredProcess.GenerateReducedSampleForClassification(testSampleSize)
monitoredProcess.PrintSample(testSample, testLabels, "testSet{0:d}.csv".format(monitoredProcess.version), "\t")
monitoredProcess.PrintStatisticsForClassification(testLabels)

